open Util
open Tokens
open Printf

(* I/O *)

let out s = print_string s

(* START OF INL STACK *)

type inltoken_sem = ITS_CITE | ITS_VAR    | ITS_CODE
type inltoken_b   = ITS_B    | ITS_STRONG | ITS_Q3
type inltoken_i   = ITS_I    | ITS_EM	  | ITS_Q2
type inltoken_s   = ITS_S    | ITS_STRIKE

type inltoken = | IT_B   of inltoken_b
		| IT_I   of inltoken_i
		| IT_U
		| IT_TT
		| IT_SEM of inltoken_sem
		| IT_SUP
		| IT_SUB
		| IT_BIG
		| IT_SMALL
		| IT_FONT of html_font
		| IT_S    of inltoken_s
		| IT_LINK of string

type inlcache = {
    mutable inlcache_sem  : inltoken_sem option;
    mutable inlcache_b    : inltoken_b   option;
    mutable inlcache_i    : inltoken_i   option;
    mutable inlcache_s    : inltoken_s   option;
    mutable inlcache_tt   : bool;
    mutable inlcache_u    : bool;
    mutable inlcache_sup  : int;
    mutable inlcache_sub  : int;
    mutable inlcache_small: int;
    mutable inlcache_big  : int;
    mutable inlcache_font : int;
    mutable inlcache_link : bool;
}

let inlcache  = {
    inlcache_sem  = None;
    inlcache_b    = None;
    inlcache_i    = None;
    inlcache_s    = None;
    inlcache_tt   = false;
    inlcache_u    = false;
    inlcache_sup  = 0;
    inlcache_sub  = 0;
    inlcache_small= 0;
    inlcache_big  = 0;
    inlcache_font = 0;
    inlcache_link = false;
}
let inlstack = ref []

let ipop () =
    match !inlstack with
	| []                  -> None
	| h::t ->
	    (match h with
	    | IT_B ITS_B      -> out "</b>";      inlcache.inlcache_b    <- None
	    | IT_B ITS_Q3     -> out "</b>";      inlcache.inlcache_b    <- None
    	    | IT_B ITS_STRONG -> out "</strong>"; inlcache.inlcache_b    <- None
	    | IT_I ITS_I      -> out "</i>";      inlcache.inlcache_i    <- None
	    | IT_I ITS_Q2     -> out "</i>";      inlcache.inlcache_i    <- None
	    | IT_I ITS_EM     -> out "</em>";     inlcache.inlcache_i    <- None
	    | IT_S ITS_S      -> out "</s>";      inlcache.inlcache_s    <- None
	    | IT_S ITS_STRIKE -> out "</strike>"; inlcache.inlcache_s    <- None
	    | IT_TT           -> out "</tt>";     inlcache.inlcache_tt   <- false
	    | IT_U            -> out "</u>";      inlcache.inlcache_u    <- false
	    | IT_SEM ITS_VAR  -> out "</var>";    inlcache.inlcache_sem  <- None
	    | IT_SEM ITS_CODE -> out "</code>";   inlcache.inlcache_sem  <- None
	    | IT_SEM ITS_CITE -> out "</cite>";   inlcache.inlcache_sem  <- None
	    | IT_BIG          -> out "</big>";    inlcache.inlcache_big  <- inlcache.inlcache_font - 1
	    | IT_SMALL        -> out "</small>";  inlcache.inlcache_small<- inlcache.inlcache_font - 1
	    | IT_SUB          -> out "</sub>";    inlcache.inlcache_sub  <- inlcache.inlcache_font - 1
	    | IT_SUP          -> out "</sup>";    inlcache.inlcache_sup  <- inlcache.inlcache_font - 1
	    | IT_FONT _       -> out "</font>";   inlcache.inlcache_font <- inlcache.inlcache_font - 1
	    | IT_LINK _       -> out "</a>";      inlcache.inlcache_link <- false;
	    ); inlstack := t; Some h

let ipush t =
    (match t with 
    | IT_B ITS_B      -> out "<b>";
			 inlcache.inlcache_b    <- Some ITS_B
    | IT_B ITS_STRONG -> out "<strong>";
			 inlcache.inlcache_b    <- Some ITS_STRONG
    | IT_B ITS_Q3     -> out "<b>";
			 inlcache.inlcache_b    <- Some ITS_Q3
    | IT_I ITS_I      -> out "<i>";
			 inlcache.inlcache_i    <- Some ITS_I
    | IT_I ITS_EM     -> out "<em>";
			 inlcache.inlcache_i    <- Some ITS_EM
    | IT_I ITS_Q2     -> out "<i>";
			 inlcache.inlcache_i    <- Some ITS_Q2
    | IT_S ITS_S      -> out "<s>";
    			 inlcache.inlcache_s    <- Some ITS_S
    | IT_S ITS_STRIKE -> out "<strike>";
    			 inlcache.inlcache_s    <- Some ITS_STRIKE
    | IT_TT           -> out "<tt>";
    			 inlcache.inlcache_tt   <- true
    | IT_U            -> out "<u>";
    			 inlcache.inlcache_u    <- true
    | IT_SEM ITS_VAR  -> out "<var>";
    			 inlcache.inlcache_sem  <- Some ITS_VAR
    | IT_SEM ITS_CODE -> out "<code>";
    			 inlcache.inlcache_sem  <- Some ITS_CODE
    | IT_SEM ITS_CITE -> out "<cite>";
    			 inlcache.inlcache_sem  <- Some ITS_CITE
    | IT_BIG          -> out "<big>";
    			 inlcache.inlcache_big  <- inlcache.inlcache_font + 1
    | IT_SMALL        -> out "<small>";
    			 inlcache.inlcache_small<- inlcache.inlcache_font + 1
    | IT_SUB          -> out "<sub>";
    			 inlcache.inlcache_sub  <- inlcache.inlcache_font + 1
    | IT_SUP          -> out "<sup>";
    			 inlcache.inlcache_sup  <- inlcache.inlcache_font + 1
    | IT_FONT a       -> let a1 = (match a.html_font_size  with None -> [] | Some v -> ("code",  v)::[]) in
			 let a2 = (match a.html_font_face  with None -> a1 | Some v -> ("face",  v)::a1) in
			 let a3 = (match a.html_font_color with None -> a2 | Some v -> ("color", v)::a2) in
			 print_htmltag "font" a1;
    			 inlcache.inlcache_font <- inlcache.inlcache_font + 1
    | IT_LINK target  -> print_htmltag "a" ["href",target];
			 inlcache.inlcache_link <- true
    ); inlstack := t :: !inlstack

let ipopx () =
    match ipop () with
    | None   -> failwith "inline state stack error"
    | Some _ -> ()

let rec izero () =
    let rec izero_aux () =
	match ipop () with
	    | None   -> ()
	    | Some _ -> izero_aux ()
    in izero_aux ()

(* END OF INL STACK *)

(* START OF PARA STACK *)

type listtype  = ListO | ListU
type tcell     = CellD | CellH
type linestate_html = LineH_CENTER
		    | LineH_H1 | LineH_H2 | LineH_H3
		    | LineH_H4 | LineH_H5 | LineH_H6
		    | LineH_P
type linestate = LineStateZ
	       | LineStateP | LineStatePre
	       | LineStateL of listtype list
	       | LineStateH1 | LineStateH2 | LineStateH3
	       | LineStateH4 | LineStateH5 | LineStateH6
	       | LineStateBQ of int
	       | LineStateHTML of linestate_html
	       | LineStateHTMLL of listtype list
               | LineStateHTMLLI of listtype list
	       | LineStateHTML_TABLE of tcell list
	       | LineStateHTML_TR    of tcell list
	       | LineStateHTML_TH    of tcell list
	       | LineStateHTML_TD    of tcell list

let rec liststart = function
    | [] 	-> ()
    | ListO::t  -> out "<ol>"; liststart t
    | ListU::t  -> out "<ul>"; liststart t
let rec listclose = function
    | []	-> ()
    | ListO::f  -> liststart f; out "</ol>";
    | ListU::f  -> liststart f; out "</ul>"
let rec listtransition = function
    | ListO::f,ListO::t -> listtransition (f,t)
    | ListU::f,ListU::t -> listtransition (f,t)
    | f,t               -> listclose f; liststart t

let parse_listtype raw =
       let typ = ref []
    in for i = String.length raw - 1 downto 0
    do
	if String.unsafe_get raw i = '#'
	    then typ := ListO :: !typ
	    else typ := ListU :: !typ
    done	    
    ; !typ

let print_transition_in_tag h =
    out (match h with
	| LineH_CENTER -> "<center>"
	| LineH_H1     -> "<h1>"
	| LineH_H2     -> "<h2>"
	| LineH_H3     -> "<h3>"
	| LineH_H4     -> "<h4>"
	| LineH_H5     -> "<h5>"
	| LineH_H6     -> "<h6>"
	| LineH_P      -> "<p>"
    )

let print_transition_out_tag h =
    out (match h with
	| LineH_CENTER -> "</center>"
	| LineH_H1     -> "</h1>"
	| LineH_H2     -> "</h2>"
	| LineH_H3     -> "</h3>"
	| LineH_H4     -> "</h4>"
	| LineH_H5     -> "</h5>"
	| LineH_H6     -> "</h6>"
	| LineH_P      -> "</p>"
    )

let ltrans (s,t) =
    if s <> t
    then begin
	(match s with 
	| LineStateZ      -> ()
	| LineStatePre    -> izero(); out "</pre>"
	| LineStateP      -> izero(); out "</p>"
	| LineStateH1     -> izero(); out "</h1>"
	| LineStateH2     -> izero(); out "</h2>"
	| LineStateH3     -> izero(); out "</h3>"
	| LineStateH4     -> izero(); out "</h4>"
	| LineStateH5     -> izero(); out "</h5>"
	| LineStateH6     -> izero(); out "</h6>"
	| LineStateHTML h -> izero(); print_transition_in_tag h
	| LineStateL f    -> izero();
			     (match t with 
			     | LineStateL t -> listtransition (f,t)
			     | _            -> listclose f
			     )
	| LineStateBQ n   -> izero();
			     (match t with 
			     | LineStateBQ m -> if n > m
			    			then for i = 1 to n-m do out "</blockquote>" done
						else if m < n
						    then for i = 1 to n-m do out "<blockquote>" done
						    else ()
			     | _             -> for i = 1 to n do out "</blockquote>" done
			     )
        | LineStateHTMLL s  -> listclose s
        | LineStateHTMLLI s -> izero(); out "</li>"; listclose s 
	);
	(match t with
	| LineStateZ      -> ()
	| LineStateP      -> out "<p>"
	| LineStatePre    -> out "<pre>"
	| LineStateH1     -> out "<h1>"
	| LineStateH2     -> out "<h2>"
	| LineStateH3     -> out "<h3>"
	| LineStateH4     -> out "<h4>"
	| LineStateH5     -> out "<h5>"
	| LineStateH6     -> out "<h6>"
	| LineStateHTML h -> print_transition_out_tag h
	| LineStateL t    -> (match s with 
			     | LineStateL _  -> ()
			     | _             -> liststart t
			     )
	| LineStateBQ n   -> (match s with 
			     | LineStateBQ _ -> ()
			     | _             -> for i = 1 to n do out "<blockquote>" done
			     )
        | LineStateHTMLL _  -> failwith "api error"
        | LineStateHTMLLI _ -> failwith "api error"
	)
    end

let line_state = ref LineStateZ

let ltransition target = ltrans (!line_state,target); line_state := target
let ltransitionh target = ltransition (LineStateHTML target)
(* END OF PARA STACK *)

let lexbuf        = Lexing.from_channel stdin
let token         = ref (Lexer.token lexbuf)
let next_token () = token := Lexer.token lexbuf
let line_start    = ref true

exception EndOfParse

let paracontrol_wiki() = (match !line_state with LineStateHTML _ -> false | _ -> true)
let doublenl_magic()   = (match !line_state with
    | LineStateHTMLL _
    | LineStateHTMLLI _
    | LineStateHTML_TABLE _
    | LineStateHTML_TR _
    | LineStateHTML_TH _
    | LineStateHTML_TD _    
        -> false
    | _ -> true)

let midline_token () =
    if !line_start && paracontrol_wiki()
    then ltransition LineStateP;
    (match !line_state with
    | LineStateHTMLL s      -> out "<li>"; line_state := LineStateHTMLLI s
    | LineStateHTML_TABLE s -> out "<tr><td>"; line_state := LineStateHTML_TD s
    | LineStateHTML_TR s    -> out "<td>";     line_state := LineStateHTML_TD s
    | _                     -> ()
    );
    line_start := false
let midline_token_print s = midline_token (); out s

let tag_open test token raw =
    midline_token ();
    if test
	then ipush token
	else out raw
let tag_close test raw =
    if test ()
	then while test () do ipopx () done
	else out raw
let tag_close2 counter raw =
    midline_token ();
    let c = counter ()
    in if c <> 0
	then while c = counter () do ipopx () done
	else out raw
let parah_open mode =
    line_start := false;
    ltransitionh mode
let parah_close mode raw =
    if !line_state = LineStateHTML mode
        then
	    begin
		ltransition LineStateZ;
	    	line_start := true;
	    end
        else midline_token_print raw

let rec parse_token () = 
    begin
    match !token with
	| EOF ->
		    ltransition LineStateZ;
		    raise EndOfParse
(* INLINE LEVEL ELEMENTS *)
	| LIT s ->
		    midline_token_print s
	| BR ->
		    midline_token_print "<br />"
	| VAR_NUMBEROFARTICLES ->
		    midline_token_print "${NUMBER_OF_ARTICLES}"
	| VAR_CURRENTTIME ->
		    midline_token_print "${CURRENT_TIME}"
	| VAR_CURRENTDAYNAME ->
		    midline_token_print "${CURRENT_DAYNAME}"
	| VAR_CURRENTYEAR ->
		    midline_token_print "${CURRENT_YEAR}"
	| VAR_CURRENTDAY ->
		    midline_token_print "${CURRENT_DAY}"
	| VAR_CURRENTMONTH ->
		    midline_token_print "${CURRENT_MONTH}"
        | LINKED(target,text) ->
                    midline_token ();
                    if inlcache.inlcache_link
                        then while inlcache.inlcache_link do ipopx () done;
                    ipush (IT_LINK target);
                    out text;
                    ipopx ()
        | T3 ->
                    midline_token ();
                    let username = "Joe R. Hacker" in
                    if inlcache.inlcache_link
                        then while inlcache.inlcache_link do ipopx () done;
                    ipush (IT_LINK ("user=" ^ username));
                    out username;
                    ipopx ();
        | T4 ->
                    midline_token ();
                    let username = "Joe R. Hacker" in
                    if inlcache.inlcache_link
                        then while inlcache.inlcache_link do ipopx () done;
                    ipush (IT_LINK ("user=" ^ username));
                    out username;
                    ipopx ();
                    out "1 Jan 1900"
	| LINK_INTERWIKI (lang,target) ->
		    midline_token ();
		    if inlcache.inlcache_link
			then while inlcache.inlcache_link do ipopx () done;
		    ipush (IT_LINK (lang ^ ":" ^ target));
		    out target;
		    ipopx ()
	| LINK target ->
		    midline_token ();
		    if inlcache.inlcache_link
			then while inlcache.inlcache_link do ipopx () done;
		    ipush (IT_LINK target)
	| LINK_CLOSE ->
		    tag_close (fun () -> inlcache.inlcache_link) "]]"
	| Q2 ->
		    midline_token ();
		    begin
			match inlcache.inlcache_i with
		        | None        ->  ipush (IT_I ITS_Q2)
			| Some ITS_Q2 ->  let q3 = inlcache.inlcache_b = Some ITS_Q3 in
					  while inlcache.inlcache_i <> None do ipopx () done;
					  if q3 && inlcache.inlcache_b = None
					    then ipush (IT_B ITS_Q3)
			| Some _      -> (while inlcache.inlcache_i <> None do ipopx () done; ipush (IT_I ITS_Q2))
		    end
	| Q3 ->
		    midline_token ();
		    begin
			match inlcache.inlcache_b with
		        | None        ->  ipush (IT_B ITS_Q3)
			| Some ITS_Q3 ->  let q2 = inlcache.inlcache_i = Some ITS_Q2 in
					  while inlcache.inlcache_b <> None do ipopx () done;
					  if q2 && inlcache.inlcache_i = None
					    then ipush (IT_I ITS_Q2)
			| Some _      -> (while inlcache.inlcache_b <> None do ipopx () done; ipush (IT_B ITS_Q3))
		    end
	| Q5   -> 
		    midline_token ();
		    begin
		    match inlcache.inlcache_i, inlcache.inlcache_b with
		    | None,        None        -> ipush (IT_B ITS_Q3); ipush (IT_I ITS_Q2)
		    | Some ITS_Q2, Some ITS_Q3 -> while (inlcache.inlcache_b <> None or inlcache.inlcache_i <> None)
						    do ipopx() done
		    | _, _                     -> failwith "Q5 magic not supported yet"
		    
		    end
	| MATH s ->
		    midline_token();
		    izero();
		    printf "<math>%s</math>" s
(* PARAGRAPH LEVEL ELEMENTS *)
	| PRE s ->
		    midline_token();
		    ltransition LineStateZ;
		    printf "<pre>%s</pre>" s
	| COLONS raw ->
		    if !line_start
		    then ltransition (LineStateBQ (String.length raw))
		    else midline_token_print raw;
	| SP s ->
		    if paracontrol_wiki()
		    then
			if !line_start
			then
			    begin
				line_start := false;
			    	ltransition LineStatePre;
				out s
			    end
			else
			    if !line_state = LineStatePre
				then out s
				else out " "
		    else out " "
	| NL ->
		    if !line_start
		    then ltransition LineStateZ
		    else
			begin
			    (match !line_state with
			    | LineStateL _   -> out "</li>"
			    | LineStatePre   -> out "\n"
			    | _              -> out " "
			    );
			end;
		    if doublenl_magic()
			then line_start := true
	| HR ->
		    ltransition LineStateZ;
		    out "<hr />";
		    line_start := true
	| LIST raw ->
		    if !line_start
		    then
			begin
			    ltransition (LineStateL (parse_listtype raw));
			    out "<li>";
			    line_start := false
			end
		    else out raw
        | EQ1 ->
                    if !line_start
                    then (ltransition LineStateH1; line_start := false)
                    else if !line_state = LineStateH1
                        then (ltransition LineStateZ; line_start := true)
                        else (midline_token (); out "=")
        | EQ2 ->
                    if !line_start
                    then (ltransition LineStateH2; line_start := false)
                    else if !line_state = LineStateH2
                        then (ltransition LineStateZ; line_start := true)
                        else (midline_token (); out "==")
        | EQ3 ->
                    if !line_start
                    then (ltransition LineStateH3; line_start := false)
                    else if !line_state = LineStateH3
                        then (ltransition LineStateZ; line_start := true)
                        else (midline_token (); out "===")
        | EQ4 ->
                    if !line_start
                    then (ltransition LineStateH4; line_start := false)
                    else if !line_state = LineStateH4
                        then (ltransition LineStateZ; line_start := true)
                        else (midline_token (); out "====")
        | EQ5 ->
                    if !line_start
                    then (ltransition LineStateH5; line_start := false)
                    else if !line_state = LineStateH5
                        then (ltransition LineStateZ; line_start := true)
                        else (midline_token (); out "=====")
        | EQ6 ->
		    if !line_start
                    then (ltransition LineStateH6; line_start := false)
                    else if !line_state = LineStateH6
                        then (ltransition LineStateZ; line_start := true)
                        else (midline_token (); out "======")

	| HTML (t,raw) ->
	begin
	match t with
	    | O_B ->		tag_open (inlcache.inlcache_b = None) (IT_B ITS_B) raw
	    | O_STRONG ->	tag_open (inlcache.inlcache_b = None) (IT_B ITS_STRONG) raw
	    | O_I ->		tag_open (inlcache.inlcache_i = None) (IT_I ITS_I) raw
	    | O_EM ->		tag_open (inlcache.inlcache_i = None) (IT_I ITS_EM) raw
	    | O_S ->		tag_open (inlcache.inlcache_s = None) (IT_S ITS_S) raw
	    | O_STRIKE ->	tag_open (inlcache.inlcache_s = None) (IT_S ITS_STRIKE) raw
	    | O_TT -> 		tag_open (not inlcache.inlcache_tt) (IT_TT) raw
	    | O_U ->		tag_open (not inlcache.inlcache_u) (IT_U) raw
	    | O_SUP ->		tag_open true (IT_SUP) raw
	    | O_SUB ->		tag_open true (IT_SUB) raw
	    | O_BIG ->		tag_open true (IT_BIG) raw
	    | O_SMALL ->	tag_open true (IT_SMALL) raw
	    | O_FONT a ->	tag_open true (IT_FONT a) raw
	    | O_VAR ->		tag_open (inlcache.inlcache_sem = None) (IT_SEM ITS_VAR) raw
	    | O_CODE ->		tag_open (inlcache.inlcache_sem = None) (IT_SEM ITS_CODE) raw
	    | O_CITE ->		tag_open (inlcache.inlcache_sem = None) (IT_SEM ITS_CITE) raw
	    | C_B ->		tag_close (fun () -> inlcache.inlcache_b = Some ITS_B) raw
	    | C_STRONG ->	tag_close (fun () -> inlcache.inlcache_b = Some ITS_STRONG) raw
	    | C_I ->		tag_close (fun () -> inlcache.inlcache_i = Some ITS_I) raw
	    | C_EM ->		tag_close (fun () -> inlcache.inlcache_i = Some ITS_EM) raw
	    | C_TT -> 		tag_close (fun () -> inlcache.inlcache_tt) raw
	    | C_U -> 		tag_close (fun () -> inlcache.inlcache_u) raw
	    | C_SUP ->		tag_close2 (fun () -> inlcache.inlcache_sup) raw
	    | C_SUB ->		tag_close2 (fun () -> inlcache.inlcache_sub) raw
	    | C_BIG ->		tag_close2 (fun () -> inlcache.inlcache_big) raw
	    | C_SMALL ->	tag_close2 (fun () -> inlcache.inlcache_small) raw
	    | C_S ->		tag_close (fun () -> inlcache.inlcache_s = Some ITS_S) raw
	    | C_STRIKE ->	tag_close (fun () -> inlcache.inlcache_s = Some ITS_STRIKE) raw
	    | C_VAR ->		tag_close (fun () -> inlcache.inlcache_sem = Some ITS_VAR) raw
	    | C_CODE ->		tag_close (fun () -> inlcache.inlcache_sem = Some ITS_CODE) raw
	    | C_CITE ->		tag_close (fun () -> inlcache.inlcache_sem = Some ITS_CITE) raw
	    | C_FONT ->		tag_close2 (fun () -> inlcache.inlcache_font) raw
	    | O_CENTER ->	parah_open  LineH_CENTER
	    | C_CENTER ->	parah_close LineH_CENTER raw
	    | O_H1 ->		parah_open  LineH_H1
	    | C_H1 ->		parah_close LineH_H1 raw
	    | O_H2 ->		parah_open  LineH_H2
	    | C_H2 ->		parah_close LineH_H2 raw
	    | O_H3 ->		parah_open  LineH_H3
	    | C_H3 ->		parah_close LineH_H3 raw
	    | O_H4 ->		parah_open  LineH_H4
	    | C_H4 ->		parah_close LineH_H4 raw
	    | O_H5 ->		parah_open  LineH_H5
	    | C_H5 ->		parah_close LineH_H5 raw
	    | O_H6 ->		parah_open  LineH_H6
	    | C_H6 ->		parah_close LineH_H6 raw
	    | O_P  ->		parah_open  LineH_P
	    | C_P  ->		parah_close LineH_P raw
            | O_UL ->
		    line_start := false;
                    (match !line_state with
                          LineStateHTMLL  s -> out "<ul>";      line_state := LineStateHTMLL (ListU::s)
                        | LineStateHTMLLI s -> out "</li><ul>"; line_state := LineStateHTMLL (ListU::s)
                        | _                 -> ltransition LineStateZ; out "<ul>";
                                               line_state := LineStateHTMLL [ListU]
                    )
            | C_UL ->
		    line_start := false;
                    (match !line_state with
                          LineStateHTMLL  (ListU::s) -> out "</ul>";      line_state := LineStateHTMLL s
                        | LineStateHTMLLI (ListU::s) -> out "</li></ul>"; line_state := LineStateHTMLL s
                        | _                          -> midline_token_print raw
                    )
            | O_OL ->
		    line_start := false;
                    (match !line_state with
                          LineStateHTMLL  s -> out "<ol>";      line_state := LineStateHTMLL (ListO::s)
                        | LineStateHTMLLI s -> out "</li><ol>"; line_state := LineStateHTMLL (ListO::s)
                        | _                 -> ltransition LineStateZ; out "<ol>";
                                               line_state := LineStateHTMLL [ListO]
                    )
            | C_OL ->
		    line_start := false;
                    (match !line_state with
                          LineStateHTMLL  (ListO::s) -> out "</ol>";      line_state := LineStateHTMLL s
                        | LineStateHTMLLI (ListO::s) -> out "</li></ol>"; line_state := LineStateHTMLL s
                        | _                          -> midline_token_print raw
                    )
            | O_LI ->
		    line_start := false;
                    (match !line_state with
                          LineStateHTMLL  s -> out "<li>"; line_state := LineStateHTMLLI s
			| LineStateHTMLLI s -> out "</li><li>"
                        | _                 -> midline_token_print raw
                    )
            | C_LI ->
		    line_start := false;
                    (match !line_state with
                          LineStateHTMLLI s -> out "</li>"; line_state := LineStateHTMLL s
                        | _                 -> midline_token_print raw
                    )
	    | O_TABLE ->
		    line_start := false;
                    (match !line_state with
                          LineStateHTML_TABLE s -> out "<tr><td><table>"; line_state := LineStateHTML_TABLE (CellD::s)
                        | LineStateHTML_TR    s -> out "<td><table>";     line_state := LineStateHTML_TABLE (CellD::s)
			| LineStateHTML_TD    s -> out "<table>";     line_state := LineStateHTML_TABLE (CellD::s)
			| LineStateHTML_TH    s -> out "<table>";     line_state := LineStateHTML_TABLE (CellH::s)
                        | _                     -> ltransition LineStateZ; out "<table>";
                                                   line_state := LineStateHTML_TABLE []
                    )
	    | C_TABLE ->
		    line_start := false;
                    (match (match !line_state with
                          LineStateHTML_TABLE s -> out "</table>"; Some s
                        | LineStateHTML_TR    s -> out "</tr></table>"; Some s
			| LineStateHTML_TD    s -> out "</td></tr></table>"; Some s
			| LineStateHTML_TH    s -> out "</th></tr></table>"; Some s
                        | _                     -> midline_token_print raw; None
                    ) with
			| None            -> ()
			| Some (CellD::s) -> line_state := LineStateHTML_TD s
			| Some (CellH::s) -> line_state := LineStateHTML_TH s
			| Some []         -> line_state := LineStateZ
		    )
	    | O_TR ->
		    line_start := false;
                    (match !line_state with
                          LineStateHTML_TABLE s -> out "<tr>"; line_state := LineStateHTML_TR s
                        | LineStateHTML_TR    s -> out "</tr><tr>"; line_state := LineStateHTML_TR s
			| LineStateHTML_TD    s -> out "</td></tr><tr>"; line_state := LineStateHTML_TR s
			| LineStateHTML_TH    s -> out "</th></tr><tr>"; line_state := LineStateHTML_TR s
                        | _                     -> midline_token_print raw
                    )
	    | C_TR -> 
		    line_start := false;
                    (match !line_state with
			| LineStateHTML_TR s -> out "</tr>";      line_state := LineStateHTML_TABLE s
			| LineStateHTML_TD s -> out "</td></tr>"; line_state := LineStateHTML_TABLE s
			| LineStateHTML_TH s -> out "</th></tr>"; line_state := LineStateHTML_TABLE s
                        | _                  -> midline_token_print raw
                    )
	    | O_TD ->
		    line_start := false;
                    (match !line_state with
                          LineStateHTML_TABLE s -> out "<tr><td>";  line_state := LineStateHTML_TD s
                        | LineStateHTML_TR    s -> out "<td>";      line_state := LineStateHTML_TD s
			| LineStateHTML_TD    s -> out "</td><td>"; line_state := LineStateHTML_TD s
			| LineStateHTML_TH    s -> out "</th><td>"; line_state := LineStateHTML_TD s
                        | _                     -> midline_token_print raw
                    )
	    | C_TD -> 
		    line_start := false;
                    (match !line_state with
			| LineStateHTML_TD    s -> out "</td>"; line_state := LineStateHTML_TR s
                        | _                     -> midline_token_print raw
                    )
	    | O_TH ->
		    line_start := false;
                    (match !line_state with
                          LineStateHTML_TABLE s -> out "<tr><th>";  line_state := LineStateHTML_TH s
                        | LineStateHTML_TR    s -> out "<th>";      line_state := LineStateHTML_TH s
			| LineStateHTML_TD    s -> out "</td><th>"; line_state := LineStateHTML_TH s
			| LineStateHTML_TH    s -> out "</th><th>"; line_state := LineStateHTML_TH s
                        | _                     -> midline_token_print raw
                    )
	    | C_TH -> 
		    line_start := false;
                    (match !line_state with
			| LineStateHTML_TH    s -> out "</th>"; line_state := LineStateHTML_TH s
                        | _                     -> midline_token_print raw
                    )
	    
	end
    end;
    next_token ();
    parse_token ()
let rec parse_headers () =
    match !token with
	| LINK_INTERWIKI (l,a) ->
		printf "<a href=\"%s:%s\">%s version</a>\n" l a l;
    		next_token ();
		parse_headers ()
	| SP s ->
    		next_token ();
		parse_headers ()
	| NL ->
    		next_token ();
		parse_headers ()
	| _ ->
		parse_token ()

let parse () = 
    try parse_headers ()
    with EndOfParse -> ()

let _ = parse ()
let _ = print_newline ()
