open Util

type html_font = {html_font_size : string option; html_font_face : string option; html_font_color : string option}
type html_tag =
	| O_FONT of html_font | C_FONT
        | O_P | O_H1 | O_H2 | O_H3 | O_H4 | O_H5 | O_H6
        | C_P | C_H1 | C_H2 | C_H3 | C_H4 | C_H5 | C_H6
        | O_UL | O_OL | O_LI | O_TABLE | O_TR | O_TH | O_TD
        | C_UL | C_OL | C_LI | C_TABLE | C_TR | C_TH | C_TD
        | O_B | O_I | O_EM | O_STRONG
        | C_B | C_I | C_EM | C_STRONG
        | O_U | O_BIG | O_SMALL | O_SUB | O_SUP
        | C_U | C_BIG | C_SMALL | C_SUB | C_SUP
        | O_CITE | O_CODE | O_S | O_STRIKE | O_TT | O_VAR
        | C_CITE | C_CODE | C_S | C_STRIKE | C_TT | C_VAR
        | O_DIV | O_CENTER | O_BLOCKQUOTE | O_CAPTION
        | C_DIV | C_CENTER | C_BLOCKQUOTE | C_CAPTION
        | O_RUBY | O_RT | O_RB | O_RP | O_DT | O_DD
        | C_RUBY | C_RT | C_RB | C_RP | C_DT | C_DD
	| Z_TR | Z_TH | Z_TD
type t  = NL | SP of string | LIST of string
	| EQ1 | EQ2 | EQ3 | EQ4 | EQ5 | EQ6
	| Q2 | Q3 | Q5
	| HR | BR
	| PRE of string
	| HTML of html_tag * string
        | MATH of string | LINK of string
	| LINK_INTERWIKI of string * string | LINK_CLOSE 
        | COLONS of string
	| T3 | T4 | LIT of string
	| LINKED of string * string
	| VAR_CURRENTMONTH | VAR_CURRENTDAY | VAR_CURRENTYEAR
    	| VAR_CURRENTDAYNAME | VAR_CURRENTTIME | VAR_NUMBEROFARTICLES
	| EOF

let char_code_0 = 48
let char_code_9 = 57
let char_code_a = 97
let char_code_z = 122
let char_code_A = 65
let char_code_Z = 90

let find_eotn str n0 =
    let rec find_eotn_aux n =
        try
            let c = Char.code (String.get str n)
            in if (c >= char_code_0 && c <= char_code_9)
               || (c >= char_code_a && c <= char_code_z)
               || (c >= char_code_A && c <= char_code_Z)
               then find_eotn_aux (n+1)
               else n
        with _ -> n
    in find_eotn_aux n0

(*
let parse_html_attrs str =
    let attrs = Hashtbl.create 0
 in let strl = String.length
 in let _ = for i = 0 to strl-1
 do
    if str.[i] = ' ' or str.[i] = '\t'
 done;
    attrs
 (* @@@ *)
*)

let parse_html_opening_tag str =
       let l     = String.length str
    in let eotn  = find_eotn str 1
    in let tn    = String.lowercase (String.sub str 1 (eotn-1))
(*
    in let attrs = (string_brange str eotn 1)
    in let _ = Printf.eprintf "@<%s[%s]>@" tn attrs
*)
    in match tn with
          "p"          -> HTML (O_P, html_escape str)
        | "h1"         -> HTML (O_H1, html_escape str)
        | "h2"         -> HTML (O_H2, html_escape str)
        | "h3"         -> HTML (O_H3, html_escape str)
        | "h4"         -> HTML (O_H4, html_escape str)
        | "h5"         -> HTML (O_H5, html_escape str)
        | "h6"         -> HTML (O_H6, html_escape str)
        | "ul"         -> HTML (O_UL, html_escape str)
        | "ol"         -> HTML (O_OL, html_escape str)
        | "li"         -> HTML (O_LI, html_escape str)
        | "table"      -> HTML (O_TABLE, html_escape str)
        | "tr"         -> HTML (O_TR, html_escape str)
        | "th"         -> HTML (O_TH, html_escape str)
        | "td"         -> HTML (O_TD, html_escape str)
        | "b"          -> HTML (O_B      ,html_escape str)
        | "i"          -> HTML (O_I      ,html_escape str)
        | "em"         -> HTML (O_EM     ,html_escape str)
        | "strong"     -> HTML (O_STRONG ,html_escape str)
        | "u"          -> HTML (O_U      ,html_escape str)
        | "big"        -> HTML (O_BIG    ,html_escape str)
        | "small"      -> HTML (O_SMALL  ,html_escape str)
        | "sub"        -> HTML (O_SUB    ,html_escape str)
        | "sup"        -> HTML (O_SUP    ,html_escape str)
        | "cite"       -> HTML (O_CITE   ,html_escape str)
        | "code"       -> HTML (O_CODE, html_escape str)
        | "s"          -> HTML (O_S, html_escape str)
        | "strike"     -> HTML (O_STRIKE, html_escape str)
        | "tt"         -> HTML (O_TT, html_escape str)
        | "var"        -> HTML (O_VAR, html_escape str)
        | "div"        -> HTML (O_DIV, html_escape str)
        | "center"     -> HTML (O_CENTER, html_escape str)
        | "blockquote" -> HTML (O_BLOCKQUOTE, html_escape str)
        | "caption"    -> HTML (O_CAPTION, html_escape str)
        | "font"       -> HTML (O_FONT {html_font_size = None; html_font_face = None; html_font_color = None}, html_escape str)
        | "ruby"       -> HTML (O_RUBY, html_escape str)
        | "rt"         -> HTML (O_RT, html_escape str)
        | "rb"         -> HTML (O_RB, html_escape str)
        | "rp"         -> HTML (O_RP, html_escape str)
        | "dt"         -> HTML (O_DT, html_escape str)
        | "dd"         -> HTML (O_DD, html_escape str)
	| "br"	       -> BR
	| "hr"	       -> HR
        | _            -> LIT (html_escape str)
	
let parse_html_closing_tag str =
       let l    = String.length str
    in let eotn = find_eotn str 2
    in let tn   = String.lowercase (String.sub str 2 (eotn-2))
    in match tn with
          "p"          -> HTML (C_P,html_escape str)
        | "h1"         -> HTML (C_H1,html_escape str)
        | "h2"         -> HTML (C_H2,html_escape str)
        | "h3"         -> HTML (C_H3,html_escape str)
        | "h4"         -> HTML (C_H4,html_escape str)
        | "h5"         -> HTML (C_H5,html_escape str)
        | "h6"         -> HTML (C_H6,html_escape str)
        | "ul"         -> HTML (C_UL,html_escape str)
        | "ol"         -> HTML (C_OL,html_escape str)
        | "li"         -> HTML (C_LI,html_escape str)
        | "table"      -> HTML (C_TABLE,html_escape str)
        | "tr"         -> HTML (C_TR,html_escape str)
        | "th"         -> HTML (C_TH,html_escape str)
        | "td"         -> HTML (C_TD,html_escape str)
        | "b"          -> HTML (C_B ,html_escape str)    
        | "i"          -> HTML (C_I ,html_escape str)    
        | "em"         -> HTML (C_EM,html_escape str)    
        | "strong"     -> HTML (C_STRONG,html_escape str)
        | "u"          -> HTML (C_U     ,html_escape str)
        | "big"        -> HTML (C_BIG   ,html_escape str)
        | "small"      -> HTML (C_SMALL ,html_escape str)
        | "sub"        -> HTML (C_SUB   ,html_escape str)
        | "sup"        -> HTML (C_SUP   ,html_escape str)
        | "cite"       -> HTML (C_CITE  ,html_escape str)
        | "code"       -> HTML (C_CODE  ,html_escape str)
        | "s"          -> HTML (C_S     ,html_escape str)
        | "strike"     -> HTML (C_STRIKE,html_escape str)
        | "tt"         -> HTML (C_TT    ,html_escape str)
        | "var"        -> HTML (C_VAR   ,html_escape str)
        | "div"        -> HTML (C_DIV,html_escape str)
        | "center"     -> HTML (C_CENTER,html_escape str)
        | "blockquote" -> HTML (C_BLOCKQUOTE,html_escape str)
        | "caption"    -> HTML (C_CAPTION,html_escape str)
        | "font"       -> HTML (C_FONT  ,html_escape str)
        | "ruby"       -> HTML (C_RUBY,html_escape str)
        | "rt"         -> HTML (C_RT,html_escape str)
        | "rb"         -> HTML (C_RB,html_escape str)
        | "rp"         -> HTML (C_RP,html_escape str)
        | "dt"         -> HTML (C_DT,html_escape str)
        | "dd"         -> HTML (C_DD,html_escape str)
        | _            -> LIT (html_escape str)
let parse_html_closed_tag str =
       let l    = String.length str
    in let eotn = find_eotn str 1
    in let tn   = String.lowercase (String.sub str 1 (eotn-1))
    in match tn with
          "br"        -> BR (* validate that attrs is empty *)
        | "hr"        -> HR (* validate that attrs is empty *)
        | "tr"        -> HTML (Z_TR,html_escape str) (* validate attrs *)
        | "th"        -> HTML (Z_TH,html_escape str) (* validate attrs *)
        | "td"        -> HTML (Z_TD,html_escape str) (* validate attrs *)
        | _           -> LIT (html_escape str)


let string_of_html_token = function 
  | Z_TR	 -> "<tr/>"
  | Z_TD	 -> "<td/>"
  | Z_TH	 -> "<th/>"
  | O_B      	 -> "<b>"
  | O_I      	 -> "<i>"
  | O_P      	 -> "<p>"
  | O_U      	 -> "<u>"
  | O_S      	 -> "<s>"
  | O_H1	 -> "<h1>"
  | O_H2	 -> "<h2>"
  | O_H3	 -> "<h3>"
  | O_H4	 -> "<h4>"
  | O_H5	 -> "<h5>"
  | O_H6	 -> "<h6>"
  | C_B      	 -> "</b>"
  | C_I     	 -> "</i>"
  | C_P      	 -> "</p>"
  | C_U      	 -> "</u>"
  | C_S      	 -> "</s>"
  | C_H1	 -> "</h1>"
  | C_H2	 -> "</h2>"
  | C_H3	 -> "</h3>"
  | C_H4	 -> "</h4>"
  | C_H5	 -> "</h5>"
  | C_H6	 -> "</h6>"
  | O_UL	 -> "<ul>"
  | O_OL	 -> "<ol>"
  | O_LI	 -> "<li>"
  | O_TR	 -> "<tr>"
  | O_TH	 -> "<th>"
  | O_TD	 -> "<td>"
  | O_EM  	 -> "<em>"
  | O_TT  	 -> "<tt>"
  | O_RT	 -> "<rt>"
  | O_RB	 -> "<rb>"
  | O_RP	 -> "<rp>"
  | O_DD	 -> "<dd>"
  | O_DT	 -> "<dl>"
  | C_UL	 -> "</ul>"
  | C_OL	 -> "</ol>"
  | C_LI	 -> "</li>"
  | C_TR	 -> "</tr>"
  | C_TH	 -> "</th>"
  | C_TD	 -> "</td>"
  | C_EM  	 -> "</em>"
  | C_TT  	 -> "</tt>"
  | C_RT	 -> "</rt>"
  | C_RB	 -> "</rb>"
  | C_RP	 -> "</rp>"
  | C_DD	 -> "</dd>"
  | C_DT	 -> "</dl>"
  | O_BIG 	 -> "<big>"
  | O_SUP 	 -> "<sup>"
  | O_SUB 	 -> "<sub>"
  | O_VAR 	 -> "<var>"
  | O_DIV	 -> "<div>"
  | C_BIG 	 -> "</big>"
  | C_SUP 	 -> "</sup>"
  | C_SUB 	 -> "</sub>"
  | C_VAR 	 -> "</var>"
  | C_DIV	 -> "</div>"
  | O_CODE 	 -> "<code>"
  | O_CITE 	 -> "<cite>"
  | O_RUBY	 -> "<ruby>"
  | C_CODE 	 -> "</code>"
  | C_CITE 	 -> "</cite>"
  | C_RUBY	 -> "</ruby>"
  | O_SMALL 	 -> "<small>"
  | O_STRIKE	 -> "<strike>"
  | O_STRONG	 -> "<strong>"
  | O_CENTER	 -> "<center>"
  | O_CAPTION	 -> "<caption>"
  | O_BLOCKQUOTE -> "<blockquote>"
  | C_SMALL 	 -> "</small>"
  | C_STRIKE 	 -> "</strike>"
  | C_STRONG 	 -> "</strong>"
  | C_CENTER	 -> "</center>"
  | C_CAPTION	 -> "</caption>"
  | C_BLOCKQUOTE -> "</blockquote>"
  | O_TABLE	 -> "<table>"
  | C_TABLE	 -> "</table>"
  | O_FONT a	 -> let a1 = (match a.html_font_size  with None -> [] | Some v -> ("code",  v)::[]) in
		    let a2 = (match a.html_font_face  with None -> a1 | Some v -> ("face",  v)::a1) in
		    let a3 = (match a.html_font_color with None -> a2 | Some v -> ("color", v)::a2) in
		    string_of_htmltag "font" a3
  | C_FONT 	-> "</font>"

let string_of_token = function
    LIT c		 -> "lit " ^ c
  | VAR_CURRENTMONTH     -> "{{CURRENTMONTH}}"
  | VAR_CURRENTDAY       -> "{{CURRENTDAY}}"
  | VAR_CURRENTYEAR      -> "{{CURRENTYEAR}}"
  | VAR_CURRENTDAYNAME   -> "{{CURRENTDAYTIME}}"
  | VAR_CURRENTTIME      -> "{{CURRENTTIME}}"
  | VAR_NUMBEROFARTICLES -> "{{NUMBEROFARTICLES}}"
  | NL       		 -> "\n"
  | SP _     		 -> "sp"
  | EOF      		 -> "eof"
  | LIST s   		 -> "list " ^ s
  | EQ1     		 -> "="
  | EQ2     		 -> "=="
  | EQ3     		 -> "==="
  | EQ4     		 -> "===="
  | EQ5     		 -> "====="
  | EQ6     		 -> "======"
  | Q2      		 -> "''"
  | Q3      		 -> "'''"
  | Q5      		 -> "'''''"
  | HR       		 -> "----"
  | PRE _    		 -> "<pre>?</pre>"
  | MATH _   		 -> "<math>?</math>"
  | LINK_INTERWIKI (w,s) -> "[[" ^ w ^ ":" ^ s ^ "]]"
  | LINK s		 -> "[[" ^ s ^ "|"
  | LINK_CLOSE           -> "]]"
  | LINKED (target,text) -> "[[" ^ target ^ "|" ^ text ^ "]]"
  | COLONS raw		 -> raw
  | T3			 -> "~~~"
  | T4			 -> "~~~~"
  | BR			 -> "<br>"
  | HTML (t,_)		 -> string_of_html_token t

let isbn_parse s = LINKED ("http://www.some.shop.com.xx/query?isbn=" ^ s, "ISBN " ^ s)
let rfc_parse  s = LINKED ("http://www.some.rfc.mirror.net.xx/" ^ s, "	RFC " ^ s)
let title_link_form s = s
let parse_ent_named code = LIT ("&" ^ code ^ ";")
let parse_ent_dec   code = LIT ("&#" ^ code ^ ";")
let parse_ent_hex   code = LIT ("&#x" ^ (String.lowercase code) ^ ";")
